/*
 * This is Tampa Ping-Pong convers/conversd derived from the wampes
 * convers package written by Dieter Deyke <deyke@mdddhd.fc.hp.com>
 * Modifications by Fred Baumgarten <dc6iq@insu1.etec.uni-karlsruhe.de>
 *
 * Modifications by Brian A. Lantz/KO4KS <brian@lantz.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "conversd.h"
#include "user.h"
#include "host.h"
#include "tnos.h"
#include "convert.h"
#include "version.h"
#include "log.h"
#include "config.h"
#include "hmalloc.h"

extern long seed;

extern struct cmdtable cmdtable[];
static int breakins;


/*---------------------------------------------------------------------------*/

void all_command(struct connection *cp)
{
	struct channel *ch;
	char *s;
	char buf[(NAMESIZE * 2) + 2];

	s = getarg(0, 1);

	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == cp->channel)
			break;
	}
	if (cp->channelop || !(ch->flags & M_CHAN_M)) {
		if (*s) {
			getTXname(cp, buf);
			send_msg_to_channel(buf, cp->channel, s);
		}
	} else {
		appendstring(cp, "*** This is a moderated channel. Only channel operators may write.\n");
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void away_command(struct connection *cp)
{
	char *s;
	char buffer[2048];
	struct connection *p;

	s = getarg(0, 1);
	if (*s) {
		sprintf(buffer, "*** (%s) You are marked as being away.\n", ts2(currtime));
		appendstring(cp, buffer);
		for (p = connections; p; p = p->next) {
			if (p->type == CT_USER) {
				if (!strcasecmp(p->name, cp->name) && !strcasecmp(p->host, cp->host)) {
					strncpy(p->away, s, AWAYSIZE);
					p->atime = currtime;
				}
			}
		}
		send_awaymsg(cp->name, myhostname, currtime, s);
	} else {
		if (*cp->away) {
			sprintf(buffer, "*** (%s) You are no longer marked as being away.\n", ts2(currtime));
			appendstring(cp, buffer);
			for (p = connections; p; p = p->next) {
				if (p->type == CT_USER) {
					if (!strcasecmp(p->name, cp->name) && !strcasecmp(p->host, cp->host)) {
						*p->away = '\0';
						p->atime = currtime;
						p->mtime = currtime;
					}
				}
			}
			send_awaymsg(cp->name, myhostname, currtime, s);
		} else {
			sprintf(buffer, "*** (%s) Actually you were marked as being here :-)\n", ts2(currtime));
			appendstring(cp, buffer);
		}
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void beep_command(struct connection *cp)
{
	if (cp->prompt[3] == '\0')
		cp->prompt[3] = '\007';
	else
		cp->prompt[3] = '\0';
	if (cp->prompt[3] == '\007')
		appendstring(cp, "*** Beep mode enabled\n");
	else
		appendstring(cp, "*** Beep mode disabled\n");
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void bye_command2(struct connection *cp, char *reason)
{
	struct connection *p;
	int users_left;
	struct clist *cl = NULLCLIST;
	int channel;

	if (!cp) {
		log(L_ERR, "OUCH, bye_command2() called with cp == NULL");
		return;
	}
	switch (cp->type) {
	case CT_UNKNOWN:
		cp->type = CT_CLOSED;
		break;
	case CT_USER:
		cp->type = CT_CLOSED;
		cl = cp->chan_list;
		while (cl) {
			users_left = count_user(cl->channel);
			if (!users_left)
				destroy_channel(cl->channel);
			clear_locks();
			send_user_change_msg(cp->name, cp->host, cl->channel, -1, reason, currtime, cp->observer);
			cp->chan_list = cl->next;
			hfree(cl);
			cl = cp->chan_list;
		}
		break;
	case CT_HOST:
		cp->type = CT_CLOSED;
		update_permlinks(cp->name, NULLCONNECTION, 0);
		for (p = connections; p; p = p->next)
			if (p->via == cp) {
				p->type = CT_CLOSED;
				clear_locks();
				channel = p->channel;
				send_user_change_msg(p->name, p->host, p->channel, -1, reason, currtime, cp->observer);
				users_left = count_user(channel);
				if (!users_left)
					destroy_channel(channel);
			}
		route_cleanup(cp);
		break;
	case CT_CLOSED:
		break;
	}
}

/*---------------------------------------------------------------------------*/

void bye_command(struct connection *cp)
{
	char *s;
	char buffer[2048];

	s = getarg(0, 1);
	if (!s || !*s) {
		bye_command2(cp, "/quit");
	} else {
		sprintf(buffer, "\"%s\"", s);
		bye_command2(cp, buffer);
	}
}


/*---------------------------------------------------------------------------*/

void new_channel(channel)
int channel;
{
	struct channel *ch;
	struct channel *chp;

	ch = (struct channel *)hmalloc(sizeof(struct channel));
	chp = channels;
	if (chp != NULLCHANNEL) {
		ch->next = NULLCHANNEL;
		while ((chp != NULLCHANNEL) && (chp->chan < channel)) {
			ch->next = chp;
			chp = chp->next;
		}
		if (ch->next == NULLCHANNEL) {
			ch->next = chp;
			channels = ch;
		} else {
			ch->next->next = ch;
			ch->next = chp;
		}
	} else {
		ch->next = NULLCHANNEL;
		channels = ch;
	}
	ch->next = chp;
	ch->chan = channel;
	ch->topic[0] = '\0';
	ch->tsetby[0] = '\0';
	ch->time = 0;
	ch->name[0] = '\0';
	ch->flags = 0;
}

/*---------------------------------------------------------------------------*/

void channel_command(struct connection *cp)
{
	char *s;
	char buffer[2048];
	char chan[64];
	int newchannel;
	struct channel *ch;
	struct connection *p;
	int printit, first;
	struct clist *cl, *cl2;

	s = getarg(0, 0);

	if (!s || !*s) {
		sprintf(buffer, "*** (%s) You are talking to channel %d. There are %d users.\n",
		    ts2(currtime), cp->channel, count_user(cp->channel));
		appendstring(cp, buffer);
		for (ch = channels; ch; ch = ch->next) {
			if (ch->chan == cp->channel)
				break;
		}
		if (ch && *ch->topic) {
			sprintf(buffer, "*** Topic set by %s: %s\n", ch->tsetby, ch->topic);
			appendstring(cp, buffer);
		}
		sprintf(buffer, "*** Also attached:");
		printit = 0;
		first = 1;
		for (cl = cp->chan_list; cl; cl = cl->next) {
			if (cl->channel != cp->channel) {
				if (first) {
					first = 0;
				} else {
					strcat(buffer, ",");
				}
				printit = count_user(cl->channel);
				if (printit == 1) {
					sprintf(chan, " channel %d (alone)", cl->channel);
				} else {
					sprintf(chan, " channel %d (%d users)", cl->channel, printit);
					printit = 1;
				}
				strcat(buffer, chan);
			}
		}
		strcat(buffer, ".\n");
		if (printit)
			appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	newchannel = atoi(s);
	if (newchannel < MINCHANNEL || newchannel > MAXCHANNEL) {
		sprintf(buffer, "*** (%s) Channel numbers must be in the range %d..%d.\n", ts2(currtime), MINCHANNEL, MAXCHANNEL);
		appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	if (newchannel == cp->channel) {
		sprintf(buffer, "*** (%s) Channel %d is already default.\n", ts2(currtime), cp->channel);
		appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == newchannel)
			break;
	}
	for (cl = cp->chan_list; cl; cl = cl->next) {
		if (cl->channel == newchannel)
			break;
	}
	if (!((cp->invitation_channel == newchannel) || !(ch) || !(ch->flags & M_CHAN_P)) && !cl) {
		sprintf(buffer, "*** (%s) You need an invitation to join the private channel %d.\n", ts2(currtime), newchannel);
		appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	if (!cl) {
		cp->locked = 1;
		send_user_change_msg(cp->name, cp->host, -1, newchannel, cp->pers, cp->time, cp->observer);
		if (!count_user(newchannel)) {
			cp->locked = 0;
			log(L_CHAN, "Channel %d created by %s@%s", newchannel, cp->name, cp->host);
		}
		cl = (struct clist *)hmalloc(sizeof(struct clist));
		cl->time = currtime;
		cp->mtime = currtime;
		if ((!cp->observer) && !ch) {
			cl->channelop = 1;
		} else {
			cl->channelop = 0;
		}
		cl->channel = newchannel;
		if (cp->chan_list->channel > newchannel) {	/* only occurs with 1st entry */
			cl->next = cp->chan_list;
			cp->chan_list = cl;
		} else
			for (cl2 = cp->chan_list; cl2; cl2 = cl2->next) {
				if (cl2->next) {	/* if there IS a next entry */
					if (cl2->next->channel > newchannel) {
						cl->next = cl2->next;
						cl2->next = cl;
						break;
					}
				} else {
					cl2->next = cl;
					cl->next = NULLCLIST;
					break;
				}
			}
	}
	cp->channel = newchannel;
	sprintf(buffer, "*** (%s) You are now talking to channel %d. ", ts2(currtime), cp->channel);
	appendstring(cp, buffer);
	if (ch) {
		if (count_user(cp->channel) == 1) {
			sprintf(buffer, "You're alone.\n");
		} else {
			sprintf(buffer, "There are %d users.\n", count_user(cp->channel));
		}
		appendstring(cp, buffer);
		if (*ch->topic) {
			sprintf(buffer, "*** Topic set by %s: %s\n", ch->tsetby, ch->topic);
			appendstring(cp, buffer);
		}
		cp->channelop = cl->channelop;
	} else {
		sprintf(buffer, "You're alone.\n");
		appendstring(cp, buffer);
		for (p = connections; p; p = p->next) {
			if (p->type == CT_HOST) {
				sprintf(buffer, "/\377\200MODE %d +o%s\n", cp->channel, cp->name);
				linklog(p, L_SENT, "%s", &buffer[3]);
				appendstring(p, buffer);
			}
		}
		if (!cp->observer)
			cp->channelop = 1;
		new_channel(cp->channel);
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void charset_command(struct connection *cp)
{

	char *s1, *s2;
	char buffer[2048];
	int charset_in, charset_out;

	s1 = getarg(0, 0);

	if (!*s1) {
		sprintf(buffer, "*** Charset in/out is %s/%s.\n",
			get_charset_by_ind(cp->charset_in),
			get_charset_by_ind(cp->charset_out));
		appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	charset_in = get_charset_by_name(s1);

	s2 = getarg(0, 0);
	if (*s2)
		charset_out = get_charset_by_name(s2);
	else
		charset_out = charset_in;

	if (charset_in < 0 || charset_out < 0) {
		sprintf(buffer, "Unknown charset: '%s'.  You may use one of them:\n%s***\n",
			(charset_in < 0) ? s1 : s2, list_charsets());
		appendstring(cp, buffer);
		appendprompt(cp, 0);
		return;
	}
	cp->charset_in = charset_in;
	cp->charset_out = charset_out;

	sprintf(buffer, "*** Charset in/out set to %s/%s.\n",
		get_charset_by_ind(cp->charset_in),
		get_charset_by_ind(cp->charset_out));
	appendstring(cp, buffer);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void cstat_command(struct connection *cp)
{
	display_links_command(cp, (*cp->name) ? cp->name : NULLCHAR);
	hosts_command(cp);
}

/*---------------------------------------------------------------------------*/

void help_command(struct connection *cp)
{
	char buf[255];
	char *search, *p;
	int prompttype = 1;

	p = search = hstrdup(getarg(0, 0));
	if (*search == '/')
		search++;
	if (!*search) {
		sprintf(buf, "Conversd Commands for %s at %s\n\n", SOFT_RELEASE, myhostname);
		appendstring(cp, buf);
	}
	if (!dohelper(cp->name, (*search) ? search : NULLCHAR)) {
		if (ecmd_exists(search)) {
			sprintf(cp->ibuf, "/\377\200HELP %s %s\n", cp->name, search);
			h_unknown_command(cp);
			sprintf(buf, "*** Extended Help Request for command '%s' sent...\n", search);
			appendstring(cp, buf);
			prompttype = 0;
		} else {
			appendstring(cp, "No help found for '");
			appendstring(cp, search);
			appendstring(cp, "'...\n");
		}
	}
	appendprompt(cp, prompttype);
	hfree(p);
}

/*---------------------------------------------------------------------------*/

void hosts_command(struct connection *cp)
{
	char buffer[2048], tmp[64];
	struct destination *d;
	int i = 1;
	char *dest;
	int prompttype = 1;

	dest = getargcs(0, 0);

	if (*dest == '\0') {
		for (d = destinations; d; d = d->next) {
			if (d->rtt) {
				sprintf(buffer, "%-9.9s (%8.8s) %3s", d->name, d->rev, ts3(d->rtt, tmp));
				appendstring(cp, buffer);
				if ((i == 3) || !(d->next)) {
					appendstring(cp, "\n");
					i = 0;
				} else {
					appendstring(cp, "   ");
				}
				i++;
			}
		}
		if ((i != 1)) {
			appendstring(cp, "\n");
		}
	} else {
		for (d = destinations; d; d = d->next) {
			if (d->rtt && !strcasecmp(d->name, dest)) {
				clear_locks();
				// check d-> link - just 2b sure..
				sprintf(buffer, "*** route: %s -> %s -> %s (%ld)\n", myhostname, ((d->link) ? d->link->name : dest), dest,  d->rtt);
				appendstring(cp, buffer);
				prompttype = 0;
				if (d->link && strcasecmp(dest, d->link->name)) {
					sprintf(buffer, "/\377\200ROUT %s %s 99\n", dest, cp->name);
					linklog(d->link->connection, L_SENT, "%s", &buffer[3]);
					appendstring(d->link->connection, buffer);
				}
				break;
			}
		}
		if (d == NULLDESTINATION) {
			sprintf(buffer, "*** no route to %s\n", dest);
			appendstring(cp, buffer);
		}
	}
	appendprompt(cp, prompttype);
}

/*---------------------------------------------------------------------------*/

void imsg_command(struct connection *cp)
{
	char *toname, *text;
	struct connection *p, *q;
	char buf[(NAMESIZE * 2) + 2];

	toname = getarg(0, 0);
	text = getarg(0, 1);
	if (!*text) {
		appendprompt(cp, 0);
		return;
	}
	for (p = connections; p; p = p->next) {
		getTXname(cp, buf);
		if (p->type == CT_USER && p->channel == cp->channel && strcasecmp(p->name, toname) != 0 && strcasecmp(p->nickname, toname) != 0) {
			send_msg_to_user(buf, p->name, text);
			q = p->next;
			while (q) {
				if (!strcasecmp(p->name, q->name)) {
					q->locked = 1;
				}
				q = q->next;
			}
		}
		if (p->via)
			p->via->locked = 0;
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void invite_command(struct connection *cp)
{
	char *toname;

	toname = getarg(0, 0);
	if (*toname)
		send_invite_msg(cp->name, toname, cp->channel);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void last_command(struct connection *cp)
{
	char *toname, *p;
	char buffer[128];
	char buffer2[128];
	FILE *f;
	int found = 0;

	toname = getarg(0, 0);
	if (!*toname) {
		appendstring(cp, "*** need username\n");
		appendprompt(cp, 0);
		return;
	}
	sprintf(buffer, "%s.%02x", userfile, get_hash(toname));
	f = fopen(buffer, "r+");
	if (f == NULL) {
		appendstring(cp, "*** can't read userfile\n");
		appendprompt(cp, 0);
		return;
	}
	while (fgets(buffer2, 512, f)) {
		if (!strncmp(buffer2, toname, strlen(toname))) {
			found++;
			for (p = buffer2; *p != ' '; p++);
			*p++ = '\0';
			for (; *p == ' '; p++);
			if (p[strlen(p) - 2] == '+') {
				p[strlen(p) - 3] = '\0';
				sprintf(buffer, "%s last sign on: %s\n", buffer2, p);
			} else {
				p[strlen(p) - 3] = '\0';
				sprintf(buffer, "%s last sign off: %s\n", buffer2, p);
			}
			appendstring(cp, buffer);
		}
	}

	if (!found) {
		sprintf(buffer, "*** %s was never logged on\n", toname);
		appendstring(cp, buffer);
	}
	fclose(f);
	appendprompt(cp, 1);
}

/*---------------------------------------------------------------------------*/

void leave_command(struct connection *cp)
{
	int chan;
	int users_left;
	struct clist *cl, *cl2;
	char *arg;
	char buffer[2048];

	arg = getarg(0, 0);
	if (*arg) {
		chan = atoi(arg);
	} else {
		chan = cp->channel;
	}
	if ((chan == cp->channel) && (cp->chan_list->next == NULLCLIST)) {
		bye_command(cp);
		return;
	}
	cl2 = cp->chan_list;
	if (cl2->channel == chan) {
		cp->chan_list = cl2->next;
		hfree(cl2);
		users_left = count_user(chan);
		if (!users_left)
			destroy_channel(chan);
		cp->locked = 1;
		send_user_change_msg(cp->name, cp->host, chan, -1, "/leave", cp->time, cp->observer);
		cp->locked = 0;
		if (chan == cp->channel) {
			cp->channel = cp->chan_list->channel;
			cp->channelop = cp->chan_list->channelop;
			sprintf(buffer, "*** (%s) Default channel is now %d.\n", ts2(currtime), cp->channel);
			appendstring(cp, buffer);
			appendprompt(cp, 0);
		} else {
			sprintf(buffer, "*** (%s) Left channel %d.\n", ts2(currtime), chan);
			appendstring(cp, buffer);
			appendprompt(cp, 0);
		}
		return;
	}
	cl = cp->chan_list;
	while (cl) {
		if (cl->channel == chan) {
			cl2->next = cl->next;
			hfree(cl);
			cl = cl2;
			users_left = count_user(chan);
			if (!users_left)
				destroy_channel(chan);
			cp->locked = 1;
			send_user_change_msg(cp->name, cp->host, chan, -1, "@", cp->time, cp->observer);
			cp->locked = 0;
			if (chan == cp->channel) {
				cp->channel = cp->chan_list->channel;
				cp->channelop = cp->chan_list->channelop;
				sprintf(buffer, "*** (%s) Default channel is now %d.\n", ts2(currtime), cp->channel);
				appendstring(cp, buffer);
				appendprompt(cp, 0);
				return;
			} else {
				sprintf(buffer, "*** (%s) Left channel %d.\n", ts2(currtime), chan);
				appendstring(cp, buffer);
				appendprompt(cp, 0);
				return;
			}
		} else {
			cl2 = cl;
		}
		cl = cl->next;
	}
	sprintf(buffer, "*** (%s) You were not on channel %d.\n", ts2(currtime), chan);
	appendstring(cp, buffer);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void display_links_command(struct connection *cp, char *user)
{
	char buffer[2048], tmp[64], tmp2[64], tmp3[64], tmp4[64];
	char *cptr;
	int pl;
	struct permlink *pp;
	long rxdivisor, txdivisor;

	if (user) {
		clear_locks();
		send_msg_to_user("conversd", user, "Host       State        Quality Software  Since NextTry Tries Queue    RX    TX");
	} else
		appendstring(cp, "Host       State        Quality Software  Since NextTry Tries Queue    RX    TX\n");
	for (pl = 0; pl < NR_PERMLINKS; pl++) {
		pp = permarray[pl];
		if (pp) {
			sprintf(buffer, "%-10.10s ", pp->name);
			cptr = &buffer[strlen(buffer)];
			if (!pp->connection) {
				strcpy(tmp, ts(pp->retrytime)),
				    sprintf(cptr, "%s   ---            %s %s  %5d",
					    (pp->primary ? "Backup link " : "Disconnected"),
				      ts(pp->statetime), tmp, pp->tries);
			} else {
				if (pp->connection->type == CT_HOST) {
					if (pp->txtime || pp->rxtime) {
						if (pp->rxtime == -1)
							sprintf(tmp2, "  %3s  ", ts3(pp->txtime, tmp3));
						else
							sprintf(tmp2, "%3s/%-3s", ts3(pp->txtime, tmp3), ts3(pp->rxtime, tmp4));
					} else
						strcpy(tmp2, "  ---  ");
					rxdivisor = (pp->connection->received > (ONE_MEG * 9)) ? ONE_MEG : ONE_K;
					txdivisor = (pp->connection->xmitted > (ONE_MEG * 9)) ? ONE_MEG : ONE_K;
					sprintf(cptr, "Connected    %s %8.8s %s               %5d %4d%c %4d%c",
						tmp2, pp->connection->rev, ts(pp->connection->time),
						queuelength(pp->connection->obuf), (int) (pp->connection->received / rxdivisor),
						(rxdivisor == ONE_K) ? 'K' : 'M', (int) (pp->connection->xmitted / txdivisor),
					(txdivisor == ONE_K) ? 'K' : 'M');
				} else {
					strcpy(tmp, ts(pp->retrytime));
					sprintf(cptr, "Connecting     ---            %s %s  %5d",
						ts(pp->statetime), tmp, pp->tries);
				}
			}
			if (user) {
				clear_locks();
				send_msg_to_user("conversd", user, buffer);
			} else {
				appendstring(cp, buffer);
				appendstring(cp, "\n");
			}
		}
	}
	if (user) {
		clear_locks();
		send_msg_to_user("conversd", user, "***");
	} else
		appendstring(cp, "\n");
}

/*---------------------------------------------------------------------------*/

void links_command(struct connection *cp)
{
	char buffer[2048];
	char *hostname, *sock_name, *cmd;
	int quality, group = 0;
	int pl;
	struct permlink *pp, *p;
	struct connection *cpp;

	hostname = getargcs(0, 0);
	cmd = getargcs(0, 0);
	if (strlen(hostname) > NAMESIZE) {
		appendstring(cp, "The host name is too long.\n");
		return;
	}
	if (!hostname || !*hostname)
		display_links_command(cp, NULLCHAR);
	else {
		if (!*cmd) {
			if (!strcasecmp(hostname, myhostname))
				display_links_command(NULLCONNECTION, cp->name);
			else
				for (cpp = connections; cpp; cpp = cpp->next) {
					if (cpp->type == CT_HOST && !cpp->locked) {
						sprintf(buffer, "/\377\200LINK %s %s\n", cp->name, hostname);
						buffer[80] = '\0';
						linklog(cpp, L_SENT, "%s", &buffer[3]);
						appendstring(cpp, buffer);
						cpp->locked = 1;
					}
				}
			return;
		}
		if (!cp->operator)
			appendstring(cp, "You must be an operator to set up new links\n");
		else {
			if (strcmp(hostname, "-")) {
				for (pl = 0; pl < NR_PERMLINKS; pl++) {
					pp = permarray[pl];
					if (pp && !strcasecmp(pp->name, hostname)) {
						if (pp->connection->type == CT_HOST)
							bye_command2(pp->connection, "link killed");
						permarray[pl] = NULLPERMLINK;
						hfree(pp);
						break;
					}
				}
				p = update_permlinks(hostname, NULLCONNECTION, 0);
				if (p) {
					group = atoi(cmd);
					quality = atoi(getarg(0, 0));
					sock_name = getarg(0, 0);
					cmd = getarg(0, 1);
					strncpy(p->name, hostname, HOSTNAMESIZE);
					p->host = hstrdup(sock_name);
					p->group = group;
					p->quality = quality;
					p->permanent = 1;
					permarray[pl] = p;
				} else
					appendstring(cp, "Link table full !\n");
			} else {
				hostname = cmd;
				for (pl = 0; pl < NR_PERMLINKS; pl++) {
					pp = permarray[pl];
					if (pp && !strcasecmp(pp->name, hostname)) {
						if (pp->connection->type == CT_HOST)
							bye_command2(pp->connection, "link killed");
						permarray[pl] = NULLPERMLINK;
						hfree(pp);
						break;
					}
				}
			}
		}
	}
	if (cp->type == CT_USER)
		appendprompt(cp, 1);
}

/*---------------------------------------------------------------------------*/

void list_command(struct connection *cp)
{
	char buffer[2048];
	char flags[16];
	struct channel *ch;
	int invisible;
	int secret;
	int special;
	int showit;
	int isonchan;
	struct connection *p;
	struct clist *cl;
	int n;
	char buf[(NAMESIZE * 2) + 2];

	appendstring(cp, "Channel Flags  Topic\n        Users\n");
	for (ch = channels; ch; ch = ch->next) {
		flags[0] = '\0';
		isonchan = 0;
		secret = 0;
		special = 0;
		invisible = 0;
		for (cl = cp->chan_list; cl; cl = cl->next) {
			if (cl->channel == ch->chan) {
				isonchan++;
			}
		}
		if (ch->flags & M_CHAN_S) {
			strcat(flags, "S");
			special++;
			secret++;
		}
		if (ch->flags & M_CHAN_P) {
			strcat(flags, "P");
		}
		if (ch->flags & M_CHAN_T) {
			strcat(flags, "T");
		}
		if (ch->flags & M_CHAN_I) {
			strcat(flags, "I");
			invisible++;
			special++;
		}
		if (ch->flags & M_CHAN_M) {
			strcat(flags, "M");
		}
		if (ch->flags & M_CHAN_L) {
			strcat(flags, "L");
		}
		showit = 0;
		if ((ch->chan == cp->channel) || isonchan || cp->operator || !special) {
			if (!ch->flags && (ch->topic[0] == '\0')) {
				sprintf(buffer, "%7d", ch->chan);
			} else {
				sprintf(buffer, "%7d %-6.6s %s\n       ", ch->chan, flags, ch->topic);
			}
			showit++;
		} else {
			if (secret && !invisible) {
				sprintf(buffer, "------- %-6.6s %s\n       ", flags, ch->topic);
				showit++;
			}
		}
		n = 9;
		if (showit) {
			for (p = connections; p; p = p->next) {
				if (n > cp->width - 12) {
					n = 9;
					strcat(buffer, "\n       ");
				}
				if (p->type == CT_USER) {
					if (p->via) {
						if (p->channel == ch->chan) {
							strcat(buffer, " ");
							getTXname(p, buf);
							strcat(buffer, buf);
							strcat(buffer, "(");
							n += 2 + strlen(p->name);
							if (!p->observer && p->operator) {
								strcat(buffer, "!");
								n++;
							} else if (!p->observer && p->channelop) {
								strcat(buffer, "@");
								n++;
							}
							if (*p->away) {
								strcat(buffer, "G");
								n++;
							}
							if (p->observer) {
								strcat(buffer, "O");
								n++;
							}
							if (buffer[strlen(buffer) - 1] == '(') {
								buffer[strlen(buffer) - 1] = '\0';
								n--;
							} else {
								strcat(buffer, ")");
							}
						}
					} else {
						for (cl = p->chan_list; cl; cl = cl->next) {
							if (cl->channel == ch->chan) {
								n += 2 + strlen(p->name);
								strcat(buffer, " ");
								getTXname(p, buf);
								strcat(buffer, buf);
								strcat(buffer, "(");
								if (!p->observer && p->operator) {
									strcat(buffer, "!");
									n++;
								} else if (!p->observer && cl->channelop) {
									strcat(buffer, "@");
									n++;
								}
								if (*p->away) {
									strcat(buffer, "G");
									n++;
								}
								if (p->observer) {
									strcat(buffer, "O");
									n++;
								}
								if (buffer[strlen(buffer) - 1] == '(') {
									buffer[strlen(buffer) - 1] = '\0';
									n--;
								} else {
									strcat(buffer, ")");
									n++;
								}
							}
						}
					}
				}
			}
			strcat(buffer, "\n");
			appendstring(cp, buffer);
		}
	}
	appendprompt(cp, 1);
}

/*---------------------------------------------------------------------------*/

void me_command(struct connection *cp)
{
	char *text;
	char buffer[2048];

	text = getarg(0, 1);
	if (*text) {
		cp->locked = 1;
		sprintf(buffer, "*** %s@%s %s", cp->name, cp->host, text);
		send_msg_to_channel("conversd", cp->channel, buffer);
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void msg_command(struct connection *cp)
{
	char *toname, *text;
	char buffer[2048];
	struct connection *p;
	int chan;
	struct channel *ch;
	struct clist *cl;
	char buf[(NAMESIZE * 2) + 2];

	toname = getarg(0, 0);
	text = getarg(0, 1);
	if (!*text) {
		appendprompt(cp, 0);
		return;
	}
	if (toname[0] == '#') {
		chan = atoi(toname + 1);
		for (cl = cp->chan_list; cl; cl = cl->next) {
			if (cl->channel == chan)
				break;
		}
		if (!cl) {
			sprintf(buffer, "*** (%s) You have not joined channel %d.\n", ts2(currtime), chan);
			appendstring(cp, buffer);
		} else {
			for (ch = channels; ch; ch = ch->next) {
				if (ch->chan == cp->channel)
					break;
			}
			if (cl->channelop || !(ch->flags & M_CHAN_M)) {
				getTXname(cp, buf);
				send_msg_to_channel(buf, cl->channel, text);
			} else {
				appendstring(cp, "*** This is a moderated channel. Only channel operators may write.\n");
			}
		}
	} else {
		for (p = connections; p; p = p->next)
			if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)))
				break;
		if (!p) {
			sprintf(buffer, "*** (%s) No such user: %s.\n", ts2(currtime), toname);
			appendstring(cp, buffer);
		} else {
			if (*p->away) {
				sprintf(buffer, "*** (%s) %s is away: %s\n", ts2(currtime), toname, p->away);
				appendstring(cp, buffer);
			}
			getTXname(cp, buf);
			send_msg_to_user(buf, toname, text);
		}
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void mode_command(struct connection *cp)
{
	char *arg, *c;
	int remove = 0;
	int channel;
	struct connection *p;
	struct channel *ch;
	int oldflags = 0;
	struct clist *cl;
	int op = 0;

	channel = atoi(getarg(0, 0));
	arg = getarg(0, 1);
	log(L_CHAN, "Mode command on channel %d: %s", channel, arg);
	for (cl = cp->chan_list; cl; cl = cl->next) {
		if ((cl->channel == channel) && (cl->channelop)) {
			op++;
			break;
		}
	}
	if (op || cp->operator || (cp->type == CT_HOST)) {
		for (ch = channels; ch; ch = ch->next) {
			if (channel == ch->chan)
				break;
		}
		if (!ch || channel == 0) {
			if (channel == 0) {
				if (cp->type == CT_USER) {
					appendstring(cp, "*** no modes on channel 0 !\n");
				}
			} else {
				appendstring(cp, "*** non existing channel !\n");
			}
		} else {
			oldflags = ch->flags;
			while (*arg) {
				switch (*arg) {
				case '+':{
						remove = 0;
						arg++;
						break;
					}
				case '-':{
						remove = 1;
						arg++;
						break;
					}
				case 'S':
				case 's':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_S);
						} else {
							ch->flags |= M_CHAN_S;
						}
						arg++;
						break;
					}
				case 'P':
				case 'p':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_P);
						} else {
							ch->flags |= M_CHAN_P;
						}
						arg++;
						break;
					}
				case 'T':
				case 't':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_T);
						} else {
							ch->flags |= M_CHAN_T;
						}
						arg++;
						break;
					}
				case 'I':
				case 'i':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_I);
						} else {
							ch->flags |= M_CHAN_I;
						}
						arg++;
						break;
					}
				case 'L':
				case 'l':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_L);
						} else {
							ch->flags |= M_CHAN_L;
						}
						arg++;
						break;
					}
				case 'M':
				case 'm':{
						if (remove && ch) {
							ch->flags -= (ch->flags & M_CHAN_M);
						} else {
							ch->flags |= M_CHAN_M;
						}
						arg++;
						break;
					}
				case 'O':
				case 'o':{
						arg++;
						while (*arg) {
							while (*arg == ' ')
								arg++;
							c = arg;
							while (*c != ' ') {
								if (*c != '\0') {
									c++;
								} else
									break;
							}
							if (*c != '\0') {
								*c++ = '\0';
							}
							for (p = connections; p; p = p->next) {
								if ((channel == -1) && (cp->type == CT_HOST)) {
									if ((!strcasecmp(p->name, arg) || !strcasecmp(p->nickname, arg)) && (p->type == CT_USER)) {
										p->operator = 1;
										send_opermsg(arg, p->host, cp->name, -1);
									}
								} else {
									if (!strcasecmp(p->name, arg) || !strcasecmp(p->nickname, arg)) {
										if (p->channel == channel) {
											p->channelop = 1;
										}
										for (cl = p->chan_list; cl; cl = cl->next) {
											if (cl->channel == p->channel)
												cl->channelop = 1;
										}
										send_opermsg(arg, p->host, cp->name, p->channel);
									}
								}
							}
							arg = c;
						}
						break;
					}
				default:{
						arg++;
						break;
					}
				}
				if (ch && (ch->flags != oldflags)) {
					if (cp->type == CT_HOST) {
						cp->locked = 1;
					}
					send_mode(ch);
				}
			}
		}
	} else {
		appendstring(cp, "*** You are not an operator !\n");
	}
	if (cp->type == CT_USER) {
		appendprompt(cp, 0);
	}
}


/*---------------------------------------------------------------------------*/

void observer_command(struct connection *cp)
{
	cp->observer = 1;
	name_command(cp);
}

/*---------------------------------------------------------------------------*/

void restricted_command(struct connection *cp)
{
	cp->type = -1;
	name_command(cp);
}

/*---------------------------------------------------------------------------*/

void fail_access(struct connection *cp)
{
	char buffer[2048];
	FILE *fd;
	
	(void) write(cp->fd, "\n*** Sorry, access restricted!\n", 31);
	
	if (!access(restrictfile, R_OK)) {
		fd = fopen(restrictfile, "r");
		if (fd != (FILE *) 0) {
			buffer[0] = '*';
			buffer[1] = ' ';
			while (!feof(fd)) {
				fgets(buffer + 2, 2045, fd);
				if (!feof(fd))
					(void) write(cp->fd, buffer, strlen(buffer));
			}
			fclose(fd);
		}
	}
}

/*---------------------------------------------------------------------------*/


void name_command(struct connection *cp)
{
	char buffer[2048];
	char usernotefile[PATH_MAX];
	int unf;
	int newchannel;
	FILE *fd;
	struct channel *ch;
	struct clist *cl;
	int cnt;
	int oldtype;
	char *p;
	
	oldtype = cp->type;
	cp->type = CT_USER;
	strncpy(cp->name, getarg(0, 0), NAMESIZE);
	if (!*cp->name) {
		appendprompt(cp, 0);
		return;
	}
	
	/* convert / to _ */
	for (p = cp->name; *p != '\0'; p++)
		if (*p == '/')
			*p = '_';
	
	/* noaccess - check for specificly denied access */
	fd = fopen(noaccessfile, "r");
	unf = 0;
	if (fd != (FILE *) 0) {
		while (!feof(fd)) {
			fgets(buffer, 2047, fd);
			rip(buffer);
			if (!strcasecmp(cp->name, buffer)) {
				unf = 1;
				break;
			}
		}
		fclose(fd);
	}
	
	if (unf) {
		log(L_AUTH, "Refused login as %s from [%s] (convers.noaccess)", cp->name, inet_ntoa(cp->sin.sin_addr));
		fail_access(cp);
		sleep(1);
		bye_command(cp);
		return;
	}
	
	/* otherwise, if restricted mode or restricted or oberserver command, check */
	if (restrictedmode || oldtype == -1) {	/* restricted access - check for permitted */
		fd = fopen(accessfile, "r");
		unf = 0;
		if (fd != (FILE *) 0) {
			while (!feof(fd)) {
				fgets(buffer, 2047, fd);
				rip(buffer);
				if (!strcasecmp(cp->name, buffer)) {
					unf = 1;
					break;
				}
			}
			fclose(fd);
		}
		if (!unf) {
			fail_access(cp);
			if ((oldtype == -1) || (restrictedmode == 1)) {		/* only valid continue */
				sleep(3);
				bye_command(cp);
				return;
			}
			/* else, let them go as observers */
			cp->observer = 1;
		}
	}
	
	/* If someone thinks they are going to play games as conversd, rudely awaken them */
	if (!strcmp(cp->name, "conversd")) {
		(void) write(cp->fd, "You are a cheater - Goodbye !\n", 30);
		sleep(3);
		bye_command(cp);
		return;
	}
	
	strncpy(cp->nickname, cp->name, NAMESIZE);
	strncpy(cp->host, myhostname, NAMESIZE);
	
	sprintf(buffer, "%s @ %s %s %s\n* Type /HELP for help.\n", convtype, myhostname, SOFT_LONGNAME, SOFT_RELEASE);
	appendstring(cp, buffer);
	
	/* print motd */
	if (!access(motdfile, R_OK)) {
		fd = fopen(motdfile, "r");
		buffer[0] = '*';
		buffer[1] = ' ';
		while (!feof(fd)) {
			fgets(buffer + 2, 2045, fd);
			if (!feof(fd)) {
				appendstring(cp, buffer);
			}
		}
		fclose(fd);
	}
	
	newchannel = atoi(getarg(0, 0));
	if (newchannel == 0)
		newchannel = defaultchannel;
	if (newchannel < MINCHANNEL || newchannel > MAXCHANNEL) {
		sprintf(buffer, "*** (%s) Channel numbers must be in the range %d..%d.\n", ts2(currtime), MINCHANNEL, MAXCHANNEL);
		appendstring(cp, buffer);
		newchannel = defaultchannel;
	}
	cp->channel = newchannel;
	log(L_LUSER, "New local user %s entering channel %d", cp->name, cp->channel);
	cnt = count_user(-1);
	sprintf(buffer, "*** There %s %d user%s online\n", (!cnt) ? "is" : "are", cnt + 1, (cnt) ? "s" : "");
	appendstring(cp, buffer);
	cnt = count_topics();
	if (cnt) {
		sprintf(buffer, "*** There %s %d topic%s available\n", (cnt == 1) ? "is" : "are", cnt, (cnt != 1) ? "s" : "");
		appendstring(cp, buffer);
	}
	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == cp->channel)
			break;
	}
	if (ch && (ch->flags & M_CHAN_P)) {
		sprintf(buffer, "*** (%s) You need an invitation to join channel %d.\n", ts2(currtime), newchannel);
		appendstring(cp, buffer);
		cp->channel = 0;
		newchannel = 0;
		for (ch = channels; ch; ch = ch->next) {
			if (ch->chan == cp->channel)
				break;
		}
	}
	if (ch) {
		cnt = count_user(newchannel);
		sprintf(buffer, "*** (%s) You are now talking to channel %d. ", ts2(currtime), newchannel);
		appendstring(cp, buffer);
		if (!cnt) {
			sprintf(buffer, "You're alone.\n");
		} else {
			sprintf(buffer, "There are %d users.\n", cnt + 1);
		}
		appendstring(cp, buffer);
		if (*ch->topic) {
			sprintf(buffer, "*** Topic set by %s: %s\n", ch->tsetby, ch->topic);
			appendstring(cp, buffer);
		}
		cp->channelop = 0;
	} else {
		sprintf(buffer, "*** You created a new channel %d.\n", cp->channel);
		appendstring(cp, buffer);
		if (!cp->observer)
			cp->channelop = 1;
		new_channel(cp->channel);
	}
	cl = (struct clist *)hmalloc(sizeof(struct clist));
	cl->next = NULLCLIST;
	cl->channelop = cp->channelop;
	cl->channel = cp->channel;
	cl->time = currtime;
	cp->mtime = currtime;
	cp->chan_list = cl;
	strcpy(cp->pers, "@");
	log(L_CHAN, "Channel %d created by %s@%s", newchannel, cp->name, cp->host);
	
	strcpy(usernotefile, DATA_DIR);
	strcat(usernotefile, "/personals/");
	strcat(usernotefile, cp->name);
	if (!access(usernotefile, R_OK)) {
		unf = open(usernotefile, O_RDONLY);
		cnt = read(unf, buffer, PERSSIZE);
		close(unf);
		if (cnt != 0) {
			strncpy(cp->pers, buffer, cnt);
			sprintf(buffer, "*** Personal data set from file: %s\n", cp->pers);
		} else {
			sprintf(buffer, "*** Failed to read personal data file\n");
		}
		appendstring(cp, buffer);
	}
	
	strcpy(usernotefile, DATA_DIR);
	strcat(usernotefile, "/nicknames/");
	strcat(usernotefile, cp->name);
	if (!access(usernotefile, R_OK)) {
		unf = open(usernotefile, O_RDONLY);
		cnt = read(unf, buffer, 512);
		close(unf);
		if (cnt != 0) {
			strncpy(cp->nickname, buffer, cnt);
			update_user_data(cp, 0);
			sprintf(buffer, "*** Nickname set from file: %s\n", cp->nickname);
		} else {
			sprintf(buffer, "*** Failed to read nickname file\n");
		}
		appendstring(cp, buffer);
	}
	send_user_change_msg(cp->name, cp->host, -1, cp->channel, cp->pers, currtime, cp->observer);
	if (!cp->observer && cp->channelop) {
		clear_locks();
		send_opermsg(cp->name, cp->host, "conversd", cp->channel);
	}
	if (cp->observer) {
		if (!access(observerfile, R_OK)) {
			fd = fopen(observerfile, "r");
			buffer[0] = '*';
			buffer[1] = ' ';
			while (!feof(fd)) {
				fgets(buffer + 2, 2045, fd);
				if (!feof(fd))
					appendstring(cp, buffer);
			}
			fclose(fd);
		} else
			appendstring(cp, "*** You are an observer, with limited commands, and no messages permitted\n");
		strcpy(buffer, OBSERVERID);
		setargp(buffer);
		personal_command(cp);
	}
}

/*---------------------------------------------------------------------------*/

void notify_command(struct connection *cp)
{
	char *p, *q, *toname;
	struct connection *pc;
	char buffer[512], t[512];
	int action;

	toname = getarg(0, 0);
	strncpy(t, toname, 512);
	toname = t;

	if (*toname == '\0') {
		sprintf(buffer, "*** (%s) You are notified if one of the following users sign on/off:\n",
			ts2(currtime));
		strcat(buffer, cp->notify);
		appendstring(cp, buffer);
		appendstring(cp, "\n");
	}
	action = 0;
	while (*toname) {
		while ((*toname == '+') || (*toname == '-')) {
			while (*toname == '+') {
				action = 0;
				toname++;
				if (*toname == '\0') {
					toname = getarg(0, 0);
					strncpy(t, toname, 512);
					toname = t;
					if (*toname == '\0') {
						break;
					}
				}
			}
			while (*toname == '-') {
				action = 1;
				toname++;
				if (*toname == '\0') {
					toname = getarg(0, 0);
					strncpy(t, toname, 512);
					toname = t;
					if (*toname == '\0') {
						break;
					}
				}
			}
		}
		strcat(toname, " ");
		p = cp->notify;
		p = strstr(p, toname);
		while (p) {
			p--;
			if (*p != ' ') {
				p++;
				p++;
			} else {
				q = ++p;
				while (*q != ' ')
					q++;
				while (*q == ' ')
					q++;
				while (*q != '\0')
					*p++ = *q++;
				*p = *q;
				p = cp->notify;
			}
			p = strstr(p, toname);
		}
		if (action == 0) {
			if (cp->notify[0] == '\0') {
				strcpy(cp->notify, " ");
			}
			strcat(cp->notify, toname);
			for (pc = connections; pc; pc = pc->next) {
				sprintf(buffer, "%s ", pc->name);
				if ((pc->type == CT_USER) && !strncmp(t, buffer, strlen(buffer))) {
					sprintf(buffer, "*** %s is online.\n", pc->name);
					appendstring(cp, buffer);
					break;
				}
			}
		}
		toname = getarg(0, 0);
		strncpy(t, toname, 512);
		toname = t;
	}
	appendprompt(cp, 0);
}


/*---------------------------------------------------------------------------*/

void operator_command(struct connection *cp)
{
	int number, answer, k;
	char buffer[2048];

	answer = atoi(getarg(0, 0));
	if (!answer) {
		if (!secretnumber) {
			sprintf(buffer, "*** Sorry, operator access is disabled at %s\n", myhostname);
			breakins++;
		} else {
			if (seed == 1L)
				conv_randomize();
			number = 0;
			for (k = 0; k < 5; k++) {
				number += conv_random(9, 0);
				if (k != 4)
					number *= 10;
			}
			cp->expected = 0;
			sprintf(buffer, "(%d) %d>\n", breakins, number);
			answer = secretnumber;
			while (answer) {
				cp->expected += (number % 10) * (answer % 10);
				number /= 10;
				answer /= 10;
			}
		}
	} else {
		if (answer == cp->expected) {
			log(L_ERR, "Operator success by %s", cp->name);
			cp->operator = 1;
			send_opermsg(cp->name, myhostname, cp->name, -1);
			sprintf(buffer, "*** (%s) You are now an operator.\n", ts2(currtime));
		} else {
			log(L_ERR, "Operator breakin by %s", cp->name);
			sprintf(buffer, "*** (%s) operator breakin attempt notified.\n", ts2(currtime));
			breakins++;
		}
	}
	appendstring(cp, buffer);
	if (answer) {
		appendprompt(cp, 0);
	}
#ifdef POSIX
	fflush(NULL);
#endif
}

/*---------------------------------------------------------------------------*/

void personal_command(struct connection *cp)
{

	char *s;
	char buffer[2048];
	char usernotefile[PATH_MAX];
	int unf;

	strcpy(usernotefile, DATA_DIR);
	strcat(usernotefile, "/personals/");
	strcat(usernotefile, cp->name);

	s = getarg(0, 1);
	if (!*s || (!strcmp(s, "@"))) {
		strcpy(s, "@");
		sprintf(buffer, "*** (%s) Personal text deleted.\n", ts2(currtime));
		unlink(usernotefile);
	} else {
		sprintf(buffer, "*** (%s) Personal text set.\n", ts2(currtime));
		if ((unf = creat(usernotefile, 0644)) != -1) {
			write(unf, s, strlen(s));
			close(unf);
		}
	}
	cp->mtime = currtime;
	appendstring(cp, buffer);
	strncpy(cp->pers, s, PERSSIZE);
	send_persmsg(cp->name, myhostname, cp->channel, s, currtime);
	update_user_data(cp, 1);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void prompt_command(struct connection *cp)
{
	char *args;

	args = getarg(0, 1);
	cp->prompt[2] = '\0';
	cp->prompt[3] = '\0';
	cp->prompt[0] = '\0';
	if ((cp->prompt[1] = args[0]) != '\0')
		if ((cp->prompt[2] = args[1]) != '\0')
			if ((cp->prompt[3] = args[2]) != '\0')
				cp->prompt[0] = args[3];
	if (cp->prompt[3] == '\b' && cp->prompt[0] == '\0') {
		cp->prompt[0] = '\b';
		cp->prompt[3] = '\0';
	}
	if (*args)
		appendstring(cp, "*** Prompting mode enabled\n");
	else
		appendstring(cp, "*** Prompting mode disabled\n");
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void query_command(struct connection *cp)
{

	char *toname;
	char buffer[2048];
	struct connection *p;

	toname = getarg(0, 0);

	if (*toname) {
		for (p = connections; p; p = p->next)
			if (p->type == CT_USER && (!strcasecmp(p->name, toname) || !strcasecmp(p->nickname, toname)))
				break;
		if (!p) {
			sprintf(buffer, "*** (%s) No such user: %s.\n", ts2(currtime), toname);
			appendstring(cp, buffer);
		} else {
			strncpy(cp->query, toname, NAMESIZE);
			sprintf(buffer, "*** (%s) Starting private conversation with %s.\n", ts2(currtime), cp->query);
			appendstring(cp, buffer);
		}
	} else if (cp->query[0] != '\0') {
		sprintf(buffer, "*** (%s) Ending private conversation with %s.\n", ts2(currtime), cp->query);
		appendstring(cp, buffer);
		cp->query[0] = '\0';
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void restart_command(struct connection *cp)
{
	int halting;


	if (!cp->operator) {
		appendstring(cp, "You must be an operator to restart/halt the system\n");
		appendprompt(cp, 1);
		return;
	}
	/* this same routine is used by both the /restart and /halt
	   commands. we now look to see if we are to halt */
	halting = (tolower(cp->ibuf[1]) == 'h');

	mysighup(0);

	if (halting) {
		/* yep, halting, so now loop indefinitely */
		for (;;);
	}
}

/*---------------------------------------------------------------------------*/

void stats_command(struct connection *cp)
{
	char *how, *howmany;
	int mode;
	long count;
	struct stats *ptr;
	int k;
	char buffer[80];
	char *tmfmtstr;
	char strbuf[20];
	int offset;
	char *fmtstr;

	how = getargcs(0, 0);
	howmany = getarg(0, 0);
	if (*how == 'd') {
		mode = 1;
		count = daysonline;
		ptr = daily;
		tmfmtstr = "%b %d: ";
		fmtstr = " %s(%9ld/%-9ld)";
	} else {		/* default to hourly */
		mode = 0;
		count = hoursonline;
		if (hoursonline > 59)
			count = 59;
		ptr = hourly;
		tmfmtstr = "%b %d, %H:00: ";
		fmtstr = " %s(%7ld/%-7ld)";
	}

	if (*howmany && (k = atoi(howmany))) {
		k--;
		if (k < count)
			count = k;
	}
	offset = (count / 2) + 1;
	sprintf(buffer, "*** %sly Statistics Profile of last %ld %s%s (rx/tx) at %s:\n",
		(mode) ? "Dai" : "Hour", count + 1, (mode) ? "day" : "hour", (count) ? "s" : "", myhostname);
	appendstring(cp, buffer);
	sprintf(buffer, "*** Online since %s***\n", ctime(&starttime));
	appendstring(cp, buffer);
	for (k = count; k >= offset; k--) {
		strftime(strbuf, 20, tmfmtstr, localtime(&ptr[k].start));
		appendstring(cp, "***  ");
		sprintf(buffer, fmtstr, strbuf, ptr[k].rx, ptr[k].tx);
		appendstring(cp, buffer);
		strftime(strbuf, 20, tmfmtstr, localtime(&ptr[k - offset].start));
		appendstring(cp, "       ");
		sprintf(buffer, fmtstr, strbuf, ptr[k - offset].rx, ptr[k - offset].tx);
		appendstring(cp, buffer);
		appendstring(cp, "\n");
	}
	if (!(count % 2)) {	/* one remaining listing */
		strftime(strbuf, 20, tmfmtstr, localtime(&ptr[k].start));
		appendstring(cp, "***  ");
		sprintf(buffer, fmtstr, strbuf, ptr[k].rx, ptr[k].tx);
		appendstring(cp, buffer);
		appendstring(cp, "\n");
	}
	appendstring(cp, "***\n");
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void sysinfo_command(struct connection *cp)
{
	char *host;
	char buffer[2048];

	host = getargcs(0, 0);
	if (!*host || !strcasecmp(host, myhostname) || !strcasecmp(host, "all")) {
		sprintf(buffer, "*** %s System Information for: %s\n*** %s: sysop mail -  %s\n*** %s: ", SOFT_TREE, myhostname, myhostname, (myemailaddr) ? myemailaddr : "*** unknown ***", myhostname);
		appendstring(cp, buffer);
		appendstring(cp, (mysysinfo) ? mysysinfo : "No additional information available");
		appendstring(cp, "\n*** ");
		appendstring(cp, (mysysinfo) ? mysyslink : "No additional informatin available");
		appendstring(cp, "\n");
		uptime_command(cp);
		if (strcasecmp(host, "all")) {
			appendprompt(cp, 0);
			return;
		}
	}
	sprintf(buffer, "/\377\200SYSI %s %s\n", cp->name, host);
	strcpy(cp->ibuf, buffer);
	h_unknown_command(cp);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void topic_command(struct connection *cp)
{
	struct channel *ch;
	struct connection *p;
	char *topic;
	char buffer[2048];
	long time;
	int channel;

	channel = cp->channel;
	topic = getarg(0, 1);
	if (*topic == '#') {
		topic++;
		while (*topic == ' ')
			topic++;
		channel = atoi(topic);
		while ((*topic != ' ') && (*topic != '\0'))
			topic++;
		while (*topic == ' ')
			topic++;
	}
	for (p = connections; p; p = p->next)
		if (p->type == CT_USER && (!strcasecmp(p->name, cp->name) || !strcasecmp(p->nickname, cp->nickname)) && (p->channel == channel))
			break;
	time = currtime;
	for (ch = channels; ch; ch = ch->next) {
		if (ch->chan == channel)
			break;
	}
	if (ch) {
		if (ch->time > time)
			time = ch->time + 1L;
		if (*topic) {
			if (p && (((ch->flags & M_CHAN_T) == 0) || cp->operator || p->channelop)) {
				if (*topic == '@') {
					*topic = '\0';
					sprintf(buffer, "*** (%s) Channel topic on channel %d removed.\n", ts2(currtime), channel);
				} else {
					sprintf(buffer, "*** (%s) Channel topic set on channel %d.\n", ts2(currtime), channel);
				}
				send_topic(cp->name, cp->host, time, channel, topic);
			} else {
				sprintf(buffer, "*** (%s) You are not an operator.\n", ts2(currtime));
			}
		} else {
			if (*ch->topic) {
				sprintf(buffer, "*** (%s) Current channel topic on channel %d, set by %s:\n    %s\n", ts2(currtime), channel, ch->tsetby, ch->topic);
			} else {
				sprintf(buffer, "*** (%s) No current channel topic on channel %d.\n", ts2(currtime), channel);
			}
		}
	} else {
		sprintf(buffer, "*** (%s) Channel channel %d non existent.\n", ts2(currtime), channel);
	}
	appendstring(cp, buffer);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void uptime_command(struct connection *cp)
{
	char buffer[2048];
#if defined(linux)
	char lx[128];
	char *chp;
	long uptime;
	FILE *fd;
#endif

	sprintf(buffer, "*** %s@%s is up for %s\n", convtype, myhostname, ts4(currtime - boottime));
#if defined(linux)
	if (!access("/proc/uptime", R_OK)) {
		appendstring(cp, buffer);
		fd = fopen("/proc/uptime", "r");
		fscanf(fd, "%ld", &uptime);
		fclose(fd);
		if (!access("/proc/version", R_OK)) {
			fd = fopen("/proc/version", "r");
			fgets(lx, 128, fd);
			fclose(fd);
			chp = lx + 15;
			while (*chp != ' ') {
				chp++;
			}
			*chp = '\0';
		} else {
			sprintf(lx, "Linux");
		}
		sprintf(buffer, "*** %s is up for %s\n", lx, ts4(uptime));
	}
#endif
	appendstring(cp, buffer);
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void verbose_command(struct connection *cp)
{
	cp->verbose = 1 - cp->verbose;
	if (cp->verbose)
		appendstring(cp, "*** Verbose mode enabled\n");
	else
		appendstring(cp, "*** Verbose mode disabled\n");
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void version_command(struct connection *cp)
{
	char buffer[2048];

	sprintf(buffer, "*** conversd %s %s\n", SOFT_LONGNAME, SOFT_RELEASE);
	appendstring(cp, buffer);
	appendstring(cp, "    This conversd implementation was originally written by Dieter Deyke\n");
	appendstring(cp, "  <deyke@mdddhd.fc.hp.com> and modified by Fred Baumgarten\n");
	appendstring(cp, "  <dc6iq@insu1.etec.uni-karlsruhe.de>.\n");

	appendstring(cp, "\n    This is a version with TNOS Conference Bridge extensions.\n");
	appendstring(cp, "  This variant (TPP) extended and maintained by Brian A. Lantz/KO4KS.\n");

	appendstring(cp, "\n    This version has also been modified by Heikki Hannikainen, OH7LZB,\n");
	appendstring(cp, "  <hessu@pspt.fi>, who removed some of Brian's features (which he\n");
	appendstring(cp, "  considered more harmful than useful) and added default channel setting\n");
	appendstring(cp, "  & optional channel 0 banning.\n***\n");
	if (*getarg(0, 0)) {
		appendstring(cp, "\nCompiler switches:\n");
		appendstring(cp, "\n");
	}
	appendprompt(cp, 0);
}

/*---------------------------------------------------------------------------*/

void who_command2(struct connection *cp)
{
	strcpy(cp->ibuf, "/who n");
	getarg(cp->ibuf, 0);
	who_command(cp);
}

/*---------------------------------------------------------------------------*/

void who_command(struct connection *cp)
{
	char buffer[2048];
	char tmp[20];
	int flags;
	int thischan = -1;
	int full = 0;
	int quick = 1;
	int aways = 0;
	int user = 0;
	char *athost = (char *) 0;
	char channelstr[16];
	char *options;
	char *cptr;
	struct connection *p;
	struct channel *ch;
	struct clist *cl;
	int showit;
	int opstat;
	int isop;
	int realname = 0;
	int usercount = 0;
	int thechannel, remaining;

	options = getarg(0, 0);
	switch (*options) {
	case '*':
		quick = 0;
		thischan = cp->channel;
		cptr = getarg(0, 0);
		if (*cptr)
			thischan = atoi(cptr);
		break;
	case 'a':
		quick = 0;
		aways = 1;
		break;
	case 'l':
		quick = 0;
		full = 1;
		break;
	case 'q':
		quick = 1;
		break;
	case 'n':
		quick = 0;
		break;
	case '@':
		athost = getargcs(0, 0);
		quick = 0;
		break;
	case 'r':
		realname = 1;
		quick = 0;
		options = getarg(0, 0);
		break;
	case 'u':
		whois_command(cp);
		return;
	}

	if (quick) {
		list_command(cp);
		return;
	}
	if (realname) {
		thechannel = 0;
		sprintf(buffer, "User             Nickname         Channel Personal\n");
		appendstring(cp, buffer);
		cp->locked = 0;
		do {
			remaining = 0;
			for (p = connections; p; p = p->next) {
				if (p->type != CT_USER || p->locked)
					continue;
				if (athost && strncasecmp(athost, p->host, strlen(athost)))
					continue;
				if (thechannel < p->channel)
					remaining = 1;
				else if (!*options || !strncasecmp(options, p->nickname, strlen(options)) || !strncasecmp(options, p->name, strlen(options))) {
					sprintf(buffer, "%-16.16s %-16.16s %7d %s\n",
						p->name, (!strcasecmp(p->name, p->nickname)) ? "" : p->nickname, p->channel, (p->pers && strcmp(p->pers, "@")) ? p->pers : "");
					appendstring(cp, buffer);
					p->locked = 1;
				}
			}
			thechannel++;
		} while (remaining);
		appendprompt(cp, 1);
		return;
	}
	appendstring(cp, "User      Host       Via      Channel  ");
	if (thischan != -1) {
		if (full) {
			appendstring(cp, " Idle Queue RX      TX\n");
		} else {
			appendstring(cp, " Idle\n");
		}
	} else {
		if (full) {
			appendstring(cp, "Login Queue RX      TX\n");
		} else {
			appendstring(cp, "Login\n");
		}
	}
	for (ch = channels; ch; ch = ch->next) {
		flags = ch->flags;
		for (p = connections; p; p = p->next) {
			showit = 0;
			opstat = 0;
			isop = 0;
			if (p->channel == ch->chan) {
				opstat = p->channelop;
				showit = 1;
			}
			isop = cp->operator;
			for (cl = cp->chan_list; cl; cl = cl->next) {
				if (ch->chan == cl->channel) {
					isop |= cl->channelop;
					break;
				}
			}
			for (cl = p->chan_list; cl; cl = cl->next) {
				if (ch->chan == cl->channel) {
					opstat = cl->channelop;
					showit = 1;
					break;
				}
			}
			if (thischan != -1 && (thischan != ch->chan))
				showit = 0;
			if (athost && showit && strncasecmp(athost, p->host, strlen(athost)))
				showit = 0;
			if (showit) {
				if (!(flags & M_CHAN_I) || (isop) || (ch->chan == cp->channel)) {
					if (!(flags & M_CHAN_S) || (isop) || (ch->chan == cp->channel)) {
						sprintf(channelstr, "%5d", ch->chan);
					} else {
						strcpy(channelstr, "-----");
					}
					if (p->type == CT_USER) {
						sprintf(buffer, full ?
							"%-8.8s%c %-10.10s %-10.10s %5s %6s %5d %7ld %7ld" :
							"%-8.8s%c %-10.10s %-10.10s %5s %6s",
							p->name, p->observer ? 'O' : p->operator ? '!' : opstat ? '@' : ' ', p->host,
							p->via ? p->via->name : "", channelstr,
							(thischan != -1) ? ts3(currtime - p->mtime, tmp) : ts(p->time),
							queuelength(p->obuf), p->received, p->xmitted);
						if (strcasecmp(p->name, p->nickname))
							if (full) {
								strcat(buffer, "\n          Nickname: ");
								strcat(buffer, p->nickname);
							}
						usercount++;
						if ((*p->pers && strcmp(p->pers, "@")) || aways) {
							if (full) {
								int n;

								n = cp->width - 21;
								strcat(buffer, "\n          Personal: ");
								strncat(buffer, p->pers, n);
							} else {
								strcat(buffer, " ");
								if (aways) {
									if (*p->away) {
										int n;

										n = cp->width - 61;
										strncat(buffer, p->away, n);

										strcat(buffer, " (since ");
										strcat(buffer, ts2(p->atime));
										strcat(buffer, ")");
									} else {
										strcat(buffer, "(here)");
									}
								} else {
									int n;

									if (*p->away) {
										n = cp->width - 53;
									} else {
										n = cp->width - 46;
									}
									strncat(buffer, p->pers, n);

								}
							}
						}
						if (!aways) {
							if (*p->away) {
								if (full) {
									int n;

									n = cp->width - 32;
									strcat(buffer, "\n          Away: ");
									strncat(buffer, p->away, n);
									strcat(buffer, " (since ");
									strcat(buffer, ts2(p->atime));
									strcat(buffer, ")");
								} else {
									strcat(buffer, " (AWAY)");
								}
							} else {
								if (full) {
									if (p->mtime) {
										strcat(buffer, "\n          Last Activity: ");
										strcat(buffer, ts2(p->mtime));
									}
								}
							}
						}
						strcat(buffer, "\n");
						if (!user || (strstr(options, p->name))) {
							appendstring(cp, buffer);
						}
					}
				}
			}
		}
	}
	if (usercount) {
		sprintf(buffer, "*** Total of %d users ***\n", usercount);
		appendstring(cp, buffer);
	} else
		appendprompt(cp, 1);
}



/*---------------------------------------------------------------------------*/

void whois_command(struct connection *cp)
{
	char buffer[2048];
	int flags;
	int aways = 0;
	char channelstr[32];
	struct connection *p;
	struct channel *ch;
	struct clist *cl;
	char *name;
	int found = 0;

	name = getarg(0, 0);
	if (!name || !*name) {
		appendstring(cp, "Need username...\n");
		appendprompt(cp, 1);
		return;
	}
	for (p = connections; p; p = p->next) {
		if ((p->type == CT_USER) && (!strcasecmp(p->name, name) || !strcasecmp(p->nickname, name))) {
			appendstring(cp, "User            Host            Via              Login Queue Receivd Xmitted\n");
			sprintf(buffer, "%-15.15s %-15.15s %-15.15s %6s %5d %7ld %7ld\n",
				p->name, p->host,
				p->via ? p->via->name : "", ts(p->time),
			  queuelength(p->obuf), p->received, p->xmitted);
			if (strcasecmp(p->name, p->nickname))
				sprintf(&buffer[strlen(buffer)], "\n      Nickname: %s", p->nickname);
			if ((*p->pers && strcmp(p->pers, "@")) || aways)
				sprintf(&buffer[strlen(buffer)], "\n      Personal: %s", p->pers);
			if (*p->away)
				sprintf(&buffer[strlen(buffer)], "\n          Away: %s (since %s)", p->away, ts2(p->atime));
			else if (p->mtime)
				sprintf(&buffer[strlen(buffer)], "\n  Last Message: %s", ts2(p->mtime));
			for (cl = p->chan_list; cl; cl = cl->next) {
				flags = 0;
				for (ch = channels; ch; ch = ch->next) {
					if (ch->chan == cl->channel) {
						flags = ch->flags;
						break;
					}
				}
				if (!cp->operator && (flags & M_CHAN_I))
					continue;
				if ((cp->operator) || !(flags & M_CHAN_S))
					sprintf(channelstr, "%-7d", ch->chan);
				else
					strcpy(channelstr, "secret ");
				if (!found) {
					sprintf(&buffer[strlen(buffer)], "\n    Channel(s): %s", channelstr);
					found = 1;
				} else
					sprintf(buffer, "\n                %s", channelstr);
				if (*ch->topic && strcmp(ch->topic, "@"))
					sprintf(&buffer[strlen(buffer)], " - %.53s", ch->topic);
			}
			if (!p->chan_list)
				sprintf(&buffer[strlen(buffer)], "\n    Channel(s): %-7d", p->channel);
			appendstring(cp, buffer);
			found = 1;
			break;
		}		/* showit */
	}			/* for connections */
	if (found)
		appendstring(cp, "\n");
	else
		appendstring(cp, "User not found...\n");
	appendprompt(cp, 1);
}



/*---------------------------------------------------------------------------*/

void width_command(struct connection *cp)
{
	int neww;
	char buffer[128];

	neww = atoi(getarg(0, 0));

	if (neww == 0) {
		sprintf(buffer, "*** Current screen width is %d\n", cp->width);
		appendstring(cp, buffer);
	} else {
		cp->width = neww;
		sprintf(buffer, "*** Screen width set to %d\n", cp->width);
		appendstring(cp, buffer);
	}
	appendprompt(cp, 0);
}

